################################################################################
### Bridge Demolition Add-on version 1.0.0
### Copyright (c) 2019 Ali Bakhtiari

# This plug-in software is provided "as is" without any express, implied warranties, or guarantees.
# In no case or event will the authors be responsible or liable for any damages or inaccuracies 
# arising from or in connection with the use of this software. 
# By using this software, you will carry the sole responsibility of any results or damages.
# Permission is granted to anyone to use the software, including commercial applications, 
# and to alter and redistribute it freely, subject to the following restrictions:
# 1.  The origin of this software must not be misrepresented. The above copyright notice appears in all copies and derivatives.
# 2.  You may freely use this software for commercial and non-commercial applications, and an acknowledgement in the product documentation would be appreciated.
# 3.  Altered source code versions must be plainly marked as such, and must not be misrepresented as being the original software.
# 4. This notice may not be removed or altered from any source distribution.

# Acknowledgement: 
# The development of this plug-in software was supported by the US Department of Transportation 
# through the Accelerated Bridge Construction University Transportation Center (ABC-UTC)
# at Florida International University as lead institution with Iowa State University and University of Nevada, Reno.
# The opinions, findings, and conclusions expressed here are those of the authors and not necessarily of the sponsor.

# Codes developed for contour visualization and renderer are based on “Deformation Visualizer v1.02 by Kai Kostack.”


################################################################################






bl_info = {
    'name': 'Bridge Demolition',
    "author": "Ali Bakhtiari",
    "version": (1, 0, 0),
    "blender": (2, 79, 0),
    "category": "All",
    "description": "Bridge Demolition Add On"
}

################################################################################
###



################################################################################


import bpy, bmesh, mathutils, sys, math, struct, time
from bpy import context
from mathutils import Vector
from mathutils import Vector
from mathutils import Color

################################################################################
# Global Variables
USING_CONTOUR = True

### Vars
qRenderAnimation = 1     
displaySteps = 200       
fadeDecrement = 2       

materialName = "ContourGradient_"

# source point of explosin
SOURCE_POINT = Vector((0,-20,7))

# Field Force Properties
FIELD_HIDE = True
FIELD_TYPE = 'FORCE'
FIELD_Z_DIRECTION = 'BOTH'
FIELD_SHAPE = 'POINT' #'PLANE'
FIELD_STRENGTH = 0.0
FIELD_FLOW = 0.0
FIELD_SEED = 40
FIELD_APPLY_TO_LOCATION = True
FIELD_APPLY_TO_ROTATION = True
FIELD_FALLOFF_TYPE = 'TUBE'
FIELD_FALLOFF_POWER = 10.0
FIELD_USE_MIN_DISTANCE = True
FIELD_USE_MAX_DISTANCE = False
FIELD_DISTANCE_MAX = 5 # 0.2
FIELD_USE_RADIAL_MIN = False
FIELD_USE_RADIAL_MAX = True
FIELD_RADIAL_MAX = 2 # 0.2
FIELD_RADIAL_FALLOFF = 0.0 # real graviational falloff = 2!!

# Rotation
ROTATION_90 = 1.5707961320877075



###############################################################################
# Panel UI
###############################################################################

class CustomDrawOperator(bpy.types.Operator):
    bl_idname = "object.custom_draw"
    bl_label = "Import"
 
    filepath = bpy.props.StringProperty(subtype="FILE_PATH")
 
    my_float = bpy.props.FloatProperty(name="Float")
    my_bool = bpy.props.BoolProperty(name="Toggle Option")
    my_string = bpy.props.StringProperty(name="String Value")
 
    def execute(self, context):
        print("filepath:", self.filepath)
        bpy.context.scene.explosion_tool.debrisfilepath = self.filepath
        return {'FINISHED'}
 
    def invoke(self, context, event):
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
 
    def draw(self, context):
        layout = self.layout
        col = layout.column()
        col.label(text="Custom Interface!")
 
        row = col.row()
        row.prop(self, "my_float")
        row.prop(self, "my_bool")
 
        col.prop(self, "my_string")


class CustomDrawOperator2(bpy.types.Operator):
    bl_idname = "object.custom_draw2"
    bl_label = "Import"
 
    filepath = bpy.props.StringProperty(subtype="FILE_PATH")
 
    my_float = bpy.props.FloatProperty(name="Float")
    my_bool = bpy.props.BoolProperty(name="Toggle Option")
    my_string = bpy.props.StringProperty(name="String Value")
 
    def execute(self, context):
        print("filepath:", self.filepath)
        bpy.context.scene.explosion_tool.logfilepath = self.filepath
        return {'FINISHED'}
 
    def invoke(self, context, event):
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
 
    def draw(self, context):
        layout = self.layout
        col = layout.column()
        col.label(text="Custom Interface!")
 
        row = col.row()
        row.prop(self, "my_float")
        row.prop(self, "my_bool")
 
        col.prop(self, "my_string")


class MessageBox(bpy.types.Operator):
    bl_idname = "message.messagebox"
    bl_label = ""
 
    message = bpy.props.StringProperty(
        name = "message",
        description = "message",
        default = ''
    )
 
    def execute(self, context):
        self.report({'INFO'}, self.message)
        print(self.message)
        return {'FINISHED'}
 
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self, width = 400)
 
    def draw(self, context):
        self.layout.label(self.message)
        self.layout.label("")




class SceneItems(bpy.types.PropertyGroup):
    source = bpy.props.FloatVectorProperty()

class AmountItems(bpy.types.PropertyGroup):
    amount = bpy.props.FloatProperty()

class DebrisLogFileOperator(bpy.types.Operator):
    bl_idname = "scene.debris_log_file_operator"
    bl_label = "Set Debris Log Filepath"

    def execute(self, context):
        bpy.ops.object.custom_draw('INVOKE_DEFAULT')
        return {'FINISHED'}

class ForceLogFileOperator(bpy.types.Operator):
    bl_idname = "scene.logfilepath_operator"
    bl_label = "Set Force Log Filepath"

    def execute(self, context):
        bpy.ops.object.custom_draw2('INVOKE_DEFAULT')
        return {'FINISHED'}


class MassButtonOperator(bpy.types.Operator):
    bl_idname = "scene.mass_button_operator"
    bl_label = "Show Total Mass of Particles"

    def execute(self, context):
        msg = 'Total Mass: ' + str(total_mass())
        bpy.ops.message.messagebox('INVOKE_DEFAULT', message = msg)
        return {'FINISHED'}


class AddButtonOperator(bpy.types.Operator):
    bl_idname = "scene.add_button_operator"
    bl_label = "Add Source Point"

    def execute(self, context):
        id = len(context.scene.collection)
        new = context.scene.collection.add()
        new.name = str(id)
        new.source = (0,0,0)
        newa = context.scene.amountcollection.add()
        newa.amount = 0
        return {'FINISHED'}

class BuildButtonOperator(bpy.types.Operator):
    bl_idname = "scene.build_button_operator"
    bl_label = "Build Force Objects"

    def execute(self, context):

        build_all_force_objects()
        return {'FINISHED'}



class ButtonOperator(bpy.types.Operator):
    bl_idname = "scene.button_operator"
    bl_label = "Button"

    id = bpy.props.StringProperty()

    def execute(self, context):
        print("Pressed button ", self.id)
        for item in context.scene.collection:
            print(item.value[0])
        return {'FINISHED'}

class DeleteOperator(bpy.types.Operator):
    bl_idname = "scene.delete_operator"
    bl_label = "Delete"

    ids = bpy.props.IntProperty()

    def execute(self, context):
        context.scene.collection.remove(self.ids)
        context.scene.amountcollection.remove(self.ids)
        return {'FINISHED'}

class FancyPanel(bpy.types.Panel):
    bl_idname = "panel.panel_fancy_panel"
    bl_label = "Explosion"
    bl_space_type = "VIEW_3D"
    bl_region_type = "TOOLS"
    bl_category = "BDA"


    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.prop(bpy.context.scene.explosion_tool, "explosion_flag")
        self.layout.operator("scene.add_button_operator")
        row = layout.row()
        i = 0
        for item in context.scene.collection:
           row = self.layout.row(align=True)
           row.prop(item, "source", text="source " + str(i))
           row.prop(context.scene.amountcollection[i], "amount")
           row.operator("scene.delete_operator", text="Delete").ids = i
           i = i + 1

        row = layout.row(align=True)
        row.operator("scene.build_button_operator")
        row = layout.row()
        self.layout.operator("scene.mass_button_operator")
        

# ------------------------------------------------------------------------
#    Scene Properties
# ------------------------------------------------------------------------

class ExplosionVariables(bpy.types.PropertyGroup):
    logfilepath = bpy.props.StringProperty(
        name = "force filepath",
        description = "write to file total contact force values"
    )
    debrisfilepath = bpy.props.StringProperty(
        name = "debris filepath",
        description = "write to file debris propagation data"
    )
    contour_flag = bpy.props.BoolProperty(
        name="Contours",
        description="Show total contact force contours",
        default=True
    )
    explosion_flag = bpy.props.BoolProperty(
        name="Explosion",
        description="Activate Explosion",
        default=False
    )
 
class ExplosionPanel(bpy.types.Panel):
    bl_idname = "panel.explosion_panel"
    bl_label = "Post Processing"
    bl_space_type = "VIEW_3D"
    bl_region_type = "TOOLS"
    bl_category = "BDA"

 
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        explosion_tool = scene.explosion_tool
        


        
        # Create an row where the buttons are aligned to each other.
        layout.label(text=" Contact Force")
        row = layout.row()
        row.prop(bpy.context.scene.explosion_tool, "contour_flag")
        row = layout.row()
        self.layout.operator("scene.debris_log_file_operator")
        self.layout.prop(bpy.context.scene.explosion_tool, 'debrisfilepath')
        
        row = layout.row()
        
        self.layout.operator("scene.logfilepath_operator")
        self.layout.prop(bpy.context.scene.explosion_tool, 'logfilepath')

class RegisterHandlerOperator(bpy.types.Operator):
    bl_idname = "scene.register_handler_operator"
    bl_label = "Register Handler"

    def execute(self, context):
        print('Initializing event handler.')
        bpy.app.handlers.frame_change_pre.append(main)
        return {'FINISHED'}

class UnregisterHandlerOperator(bpy.types.Operator):
    bl_idname = "scene.unregister_handler_operator"
    bl_label = "Unregister Handler"

    def execute(self, context):
        print('Removing event handler.')
        for i in range( len( bpy.app.handlers.frame_change_pre ) ):
            bpy.app.handlers.frame_change_pre.pop()
        return {'FINISHED'}

class HandlerPanel(bpy.types.Panel):
    bl_idname = "panel.handler_panel"
    bl_label = "Frame Change Event Handler"
    bl_space_type = "VIEW_3D"
    bl_region_type = "TOOLS"
    bl_category = "BDA"
 
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        
        self.layout.operator("scene.register_handler_operator")
        self.layout.operator("scene.unregister_handler_operator")



class ExplosionOperator (bpy.types.Operator):
    bl_idname = "wm.explosion_operator"
    bl_label = "Add Source Point"

    def execute(self, context):
        scene = context.scene
        explosion_tool = scene.explosion_tool

        i = 0
        for item in context.scene.collection:
           print(item.source[0], item.source[1], item.source[2], context.scene.amountcollection[i].amount)
           i = i + 1
        return {'FINISHED'}


###############################################################################
# Contour & Explosion
###############################################################################



def total_mass():
        total_mass = 0.0
        for obj in bpy.data.objects:
            if obj.type != 'MESH':
                continue

            if not hasattr(obj.rigid_body, 'type'):
                continue

            if obj.rigid_body.type != 'ACTIVE':
                continue
            
            # create_empty_obj
            total_mass = total_mass + obj.rigid_body.mass

        return total_mass

###############################################################################

def log_debris_locations():
    if (not bpy.context.scene.explosion_tool.debrisfilepath):
        bpy.context.scene.explosion_tool.debrisfilepath = "debrisfilepath.log"
    f = open(bpy.context.scene.explosion_tool.debrisfilepath, "a+")
    for obj in bpy.data.objects:
        if obj.type != 'MESH':
                continue
        if not hasattr(obj.rigid_body, 'type'):
            continue

        if obj.rigid_body.type != 'ACTIVE':
            continue
        f.write(obj.name + ":" + str(obj.rigid_body.location[0]) + "," + str(obj.rigid_body.location[1]) + "," + str(obj.rigid_body.location[2]) + "\n")

    f.close()


###############################################################################

def build_all_force_objects():
    for obj in bpy.data.objects:
        if obj.type != 'MESH':
            continue

        if not hasattr(obj.rigid_body, 'type'):
            continue

        if obj.rigid_body.type != 'ACTIVE':
            continue
        
        # create_empty_obj
        create_empty_obj(obj, 0)
        create_empty_obj(obj, 1)
        create_empty_obj(obj, 2)

################################################################################
# create empty object

def create_empty_obj(parent_obj, axes_index):
    emptyname = 'Empty_' + str(axes_index) + '_' + parent_obj.name
    
    if bpy.data.objects.get(emptyname) is None:
        bpy.ops.object.empty_add(type='PLAIN_AXES')
        emptyobj = bpy.data.objects[bpy.context.active_object.name]
        emptyobj.name = emptyname
        emptyobj.select = False
        emptyobj.hide = FIELD_HIDE
        #emptyobj.parent = parent_obj
        emptyobj.location = parent_obj.location
        emptyobj.rotation_euler[0] = 0
        emptyobj.rotation_euler[1] = 0
        emptyobj.rotation_euler[2] = 0
        if (axes_index < 2):
            emptyobj.rotation_euler[axes_index] = ROTATION_90

        emptyobj.field.type = FIELD_TYPE
        emptyobj.field.z_direction = FIELD_Z_DIRECTION
        emptyobj.field.shape = FIELD_SHAPE
        emptyobj.field.strength = FIELD_STRENGTH
        emptyobj.field.flow = FIELD_FLOW
        emptyobj.field.seed = FIELD_SEED
        emptyobj.field.apply_to_location = FIELD_APPLY_TO_LOCATION
        emptyobj.field.apply_to_rotation = FIELD_APPLY_TO_ROTATION
        emptyobj.field.falloff_type = FIELD_FALLOFF_TYPE
        emptyobj.field.falloff_power = FIELD_FALLOFF_POWER
        emptyobj.field.use_min_distance = FIELD_USE_MIN_DISTANCE
        emptyobj.field.use_max_distance = FIELD_USE_MAX_DISTANCE
        emptyobj.field.distance_max = FIELD_DISTANCE_MAX
        emptyobj.field.use_radial_min = FIELD_USE_RADIAL_MIN
        emptyobj.field.use_radial_max = FIELD_USE_RADIAL_MAX
        emptyobj.field.radial_max = FIELD_RADIAL_MAX
        emptyobj.field.radial_falloff = FIELD_RADIAL_FALLOFF


################################################################################
def update_empty_objects(parent_obj, explosion_force):
    for i in range(3):
        emptyobj = bpy.data.objects["Empty_" + str(i) + "_" + parent_obj.name]
        emptyobj.location = parent_obj.location
        #emptyobj.field.strength = explosion_force[i]
        if i < 2:
            emptyobj.field.strength = explosion_force[abs(i-1)]
        else :
            emptyobj.field.strength = explosion_force[i]


################################################################################

# explosion

def explosion(m_origin, m_source, time_elapsed, projected_area, amount):
    Distance = [0,0,0] #btVector3

    w = amount;

    Distance[0] = m_origin[0] - m_source[0]
    Distance[1] = m_origin[1] - m_source[1]
    Distance[2] = m_origin[2] - m_source[2]

    distance = pow((Distance[0] * Distance[0] + Distance[1] * Distance[1] + Distance[2] * Distance[2]), 0.5);

    distance /= pow(w,1/3);


    area = projected_area;

    ambientPressure = 101325.0;

    peakOverpressure = ambientPressure * 808 * ((pow(distance / 4.5, 2) + 1) / (pow((1 + (pow(distance / 0.048, 2))), 0.5) * pow((1 + (pow(distance / 0.32, 2))), 0.5) * pow((1 + (pow(distance / 1.35, 2))), 0.5)));

    millisecond_to_second = 0.001

    td = millisecond_to_second * 980 * (1 + pow(distance / 0.54, 10) / ((1 + pow(distance / 0.02, 3.0)) * (1 + pow(distance / 0.74, 6)) * (pow(1 + pow(distance / 6.9, 2), 0.5))) * pow(w, 1/3));

    beta = -1.0;

    #Friedlander Equation
    pressure = ambientPressure + peakOverpressure * (1 - time_elapsed / td) * math.exp(beta * time_elapsed / td);

    explosionforcex = pressure * area * Distance[0] / pow((Distance[0] * Distance[0] + Distance[1] * Distance[1] + Distance[2] * Distance[2]), 0.5);
    explosionforcey = pressure * area * Distance[1] / pow((Distance[0] * Distance[0] + Distance[1] * Distance[1] + Distance[2] * Distance[2]), 0.5);
    explosionforcez = pressure * area * Distance[2] / pow((Distance[0] * Distance[0] + Distance[1] * Distance[1] + Distance[2] * Distance[2]), 0.5);

    m_explosionForce = [explosionforcex, explosionforcey, explosionforcez];
    return m_explosionForce;



################################################################################
# bmesh from object

def bmesh_from_object(ob):
    matrix = ob.matrix_world
    me = ob.to_mesh(bpy.data.scenes['Scene'], apply_modifiers=True, settings='PREVIEW')
    me.transform(matrix)
    bm = bmesh.new()
    bm.from_mesh(me)
    bpy.data.meshes.remove(me)
    return bm

################################################################################
# dot product

def dot_product(vec1, vec2):
    sum = vec1[0]*vec2[0]+vec1[1]*vec2[1]+vec1[2]*vec2[2]
    return sum

################################################################################
# projected area

def projected_area(obj, projection_normal_vector):
    bm = bmesh_from_object(obj)
    bm.normal_update()
    print('bmesh:', bm)

    pa = 0
    for f in bm.faces:
        face_area = f.calc_area()
        face_normal = f.normal
        
        # compute projected area
        face_pa = abs(dot_product(projection_normal_vector, face_normal)) * face_area
        pa += face_pa
        print('face index:', f.index, ' and face area:', face_area, 'face normal:', face_normal, 'face pa:', face_pa)
        print

    bm.free()
    pa = pa/2
    return pa




################################################################################

def do_explosion(current_frame):
    print("do:", current_frame)
    # deselect all of the objects
    bpy.ops.object.select_all(action='DESELECT')

    time_scale = bpy.context.scene.rigidbody_world.time_scale
    steps_per_second = bpy.context.scene.rigidbody_world.steps_per_second

    time_elapsed = current_frame * time_scale / steps_per_second

    if (not bpy.context.scene.explosion_tool.logfilepath):
        bpy.context.scene.explosion_tool.logfilepath = "forcelogfile_explosion.log"
    f = open(bpy.context.scene.explosion_tool.logfilepath+"_explosion.log", "a+")
    f.write('name, projected area, explosion force, projection_normal_vector\n')
    f.write('current_frame: ' + str(current_frame) + 'elapsed time: ' + str(time_elapsed) + '\n')
    for obj in bpy.data.objects:
        if obj.type != 'MESH':
            continue

        if not hasattr(obj.rigid_body, 'type'):
            continue

        if obj.rigid_body.type != 'ACTIVE':
            continue

        projection_normal_vector = obj.rigid_body.location - SOURCE_POINT

        pa = projected_area(obj, projection_normal_vector)

        index = 0
        total_explosion_force = [0, 0, 0]
        for item in bpy.context.scene.collection:
            amount = bpy.context.scene.amountcollection[index].amount
            print("amount", amount)
            explosion_force = explosion(obj.rigid_body.location, item.source, time_elapsed, pa, amount)
            print("explosion_force: ", explosion_force)
            for i in range(3):
                total_explosion_force[i] = total_explosion_force[i] + explosion_force[i]
                print("total_explosion_force: ", total_explosion_force)
            index = index +1

        print('name:', obj.name, 'projected area:', pa, 'explosion force:', total_explosion_force, 'projection_normal_vector:', projection_normal_vector)
        f.write(obj.name + ',' + str(pa) + ',' + str(explosion_force) + ',' + str(projection_normal_vector) + '\n')

        # apply force
        update_empty_objects(obj, total_explosion_force)
 
    f.close()

################################################################################

def main(scene):
    

    ### Check if last frame is reached then stop animation playback
    if scene.frame_current == scene.frame_end:
        log_debris_locations()
        if bpy.context.screen.is_animation_playing:
            bpy.ops.screen.animation_play()
            

    if not "start-frame" in bpy.app.driver_namespace:
        em = bpy.app.driver_namespace["start-frame"] = [] 
        log_debris_locations()
    

    #############################
    ### What to do on start frame
    
    if bpy.context.scene.explosion_tool.contour_flag:
            
            ### Render part
            if qRenderAnimation:
                # need to disable handler while rendering, otherwise Blender crashes
                for i in range( len( bpy.app.handlers.frame_change_pre ) ):
                     bpy.app.handlers.frame_change_pre.pop()
                
                filepathOld = bpy.context.scene.render.filepath
                bpy.context.scene.render.filepath += "%04d" %(scene.frame_current -1)
                bpy.context.scene.render.image_settings.file_format = 'JPEG'
                bpy.context.scene.render.image_settings.quality = 75
                
                # Render single frame with render settings
                if qRenderAnimation == 1: bpy.ops.render.render(write_still=True)
                # Render single frame in OpenGL mode
                elif qRenderAnimation == 2: bpy.ops.render.opengl(write_still=True)
                
                bpy.context.scene.render.filepath = filepathOld
                
                # Append handler again
                bpy.app.handlers.frame_change_pre.append(main)


                #############################
                ### What to do on start frame
                if not "force-contour" in bpy.app.driver_namespace:
                    print("Initializing buffers...")
                    
                    
                    ###### Function
                    initBuffers(scene)
                    ###### Function
                    initMaterials()
                else:
                    ###### Function
                    ################################
                    ### What to do AFTER start frame
                    print("Frame:", scene.frame_current)
                    
                    defObjs = bpy.app.driver_namespace["force-contour"]
                     
                    ###### Function
                    changeObjMaterials()
                        

    
    if bpy.context.scene.explosion_tool.explosion_flag:
        if not "explosion" in bpy.app.driver_namespace:
            print("Initializing buffers...")
            ###### Function
            initExplosion(scene)
        else:
            
            defObjsExplosion = bpy.app.driver_namespace["explosion"]
            do_explosion(scene.frame_current)


    ################################
    ### What to do AFTER start frame
    print("Frame:", scene.frame_current)
        
            



################################################################################

def initExplosion(scene):
    # Create property for frame change storage of data
    defObjsExplosion = bpy.app.driver_namespace["explosion"] = [] 
    

                            
################################################################################

def initBuffers(scene):
    
    
    
    ### Make object list
    objs = []
    for obj in bpy.data.objects:
        # exclude constraint objects
        if obj.select and obj.type == 'MESH' and not obj.hide and obj.is_visible(scene) and not hasattr(obj.rigid_body_constraint, 'type'):
            objs.append(obj)
    
    # Create property for frame change storage of data
    defObjs = bpy.app.driver_namespace["force-contour"] = [] 
    defObjsDict = bpy.app.driver_namespace["force-contour-dict"] = {}

    # frame force dictionary
    forceDict = {}

    # for all constraints calculate object forces
    objects = bpy.data.objects
    for obj in objects:
        if(hasattr(obj.rigid_body_constraint, 'type')):

            if obj.rigid_body_constraint.object1.name not in forceDict:
                forceDict[obj.rigid_body_constraint.object1.name] = 0

            if obj.rigid_body_constraint.object2.name not in forceDict:
                forceDict[obj.rigid_body_constraint.object2.name] = 0
    

    ### Create original transform data array
    d = -1
    for objname in forceDict:
        d += 1
        if d %100 == 0: sys.stdout.write('\r' +"%d" %d)
        defObjForceDif = 0
        defObjs.append([obj, forceDict[objname], defObjForceDif])
        defObjsDict[objname] = d
        ### Remove all materials but one from object
        objectd = bpy.data.objects[objname]
        bpy.context.scene.objects.active = objectd

        while len(objectd.material_slots) > 1:
            objectd.active_material_index = 0
            bpy.ops.object.material_slot_remove()
            
        # Add property for the fading effect
        objectd["defLastStep"] = 0
        
    print("defObjsDict", defObjsDict)
        
################################################################################

def initMaterials():

    
    
    ### Create gradient materials for later use and reuse
    for step in range(displaySteps):
        dif = step *(1 /(displaySteps -1))
        
        col = Color((0, 0, 0))
        #col.h = 0; col.s = 1; col.v = 1
        difNormal = dif
        col.r = difNormal   # correct math: = (difNormal -0.5) *2
        col.g = 1 -(abs(0.5 -difNormal) *2)
        col.b = (0.5 -difNormal) *2
        
        mat = bpy.data.materials.new(materialName +"%03d" %step)
        mat.diffuse_color = col
        col.s *= 0.5
        mat.specular_color = col
        mat.emit = 0.33


################################################################################


def changeObjMaterials():

    
    
    time_scale = bpy.context.scene.rigidbody_world.time_scale
    steps_per_second = bpy.context.scene.rigidbody_world.steps_per_second

    # Get property data
    defObjs = bpy.app.driver_namespace["force-contour"] 
    defObjsDict = bpy.app.driver_namespace["force-contour-dict"]

    # frame force dictionary
    forceDict = {}
    # for all constraints calculate object forces
    objects = bpy.data.objects
    for obj in objects:
        if(hasattr(obj.rigid_body_constraint, 'type')):
            impulse = obj.rigid_body_constraint.appliedImpulse()
            force = impulse / time_scale * steps_per_second

            if obj.rigid_body_constraint.object1.name in forceDict:
                forceDict[obj.rigid_body_constraint.object1.name] += force
            else:
                forceDict[obj.rigid_body_constraint.object1.name] = force

            if obj.rigid_body_constraint.object2.name in forceDict:
                forceDict[obj.rigid_body_constraint.object2.name] -= force
            else:
                forceDict[obj.rigid_body_constraint.object2.name] = 0 - force
    


    if (not bpy.context.scene.explosion_tool.logfilepath):
        bpy.context.scene.explosion_tool.logfilepath = "forcelogfile.log"
    f = open(bpy.context.scene.explosion_tool.logfilepath, "a+")
    listForces = []
    for defObj in forceDict:
        f.write(str(defObj) + ":" + str(forceDict[defObj]) + "\n")
        listForces.append(forceDict[defObj])

    minf = 0.0
    maxf = 0.0

    for li in listForces:
        if li < minf:
            minf = li

        if li > maxf:
            maxf = li

    
   
    d = -1
    for defObj in forceDict:
        d += 1
        if d %100 == 0: sys.stdout.write('\r' +"%d" %d)
        objectd = bpy.data.objects[defObj]
        defObjForce = defObjs[defObjsDict[defObj]][1] # old force
        
        # print("old force: ", defObjForce)
        # f.write("old force" + str(defObjForce) + ",")

        ### Find material by deformation
        objForce = forceDict[defObj]   # new force
        
        # print("new force: ", objForce)
        # f.write("new force: " + str(objForce) + ",")

        # Calculate deformation differences
        defObjForceDif = defObjForce - objForce
        
        defObjs[defObjsDict[defObj]][1] = objForce # set new force
        defObjs[defObjsDict[defObj]][2] = defObjForceDif # set new forceDif
       
        # print("defObjForceDif: ", defObjForceDif)
        # f.write("defObjForceDif: " + str(defObjForceDif) + ",")

        # Use the deformation type with the greatest amplitude
        dif = abs(defObjForceDif)

        diff = dif /2000
        # Only calculate gradient mat index when the actual normalized deformation lies within [0..1] otherwise use the last gradient material
        if diff < displaySteps: step = int(diff)
        else: step = displaySteps -1
        
        
        # Calculate fading effect
        stepLast = objectd["defLastStep"]
        if step < stepLast -fadeDecrement: step = stepLast -fadeDecrement
        
        mat = bpy.data.materials[materialName +"%03d" %step]
        objectd["defLastStep"] = step
        
        ### add new materials slot and material

        objectd.material_slots[-1:][0].material = mat
    
    
                

################################################################################

# Registration
###############################################################################

def register():

    bpy.utils.register_class(SceneItems)
    bpy.utils.register_class(AmountItems)
    bpy.utils.register_class(ForceLogFileOperator)
    bpy.utils.register_class(DebrisLogFileOperator)
    bpy.utils.register_class(AddButtonOperator)
    bpy.utils.register_class(MassButtonOperator)
    bpy.utils.register_class(BuildButtonOperator)
    bpy.utils.register_class(ButtonOperator)
    bpy.utils.register_class(DeleteOperator)
    bpy.types.Scene.collection = bpy.props.CollectionProperty(type=SceneItems)
    bpy.types.Scene.amountcollection = bpy.props.CollectionProperty(type=AmountItems)
    bpy.utils.register_class(CustomDrawOperator2)
    bpy.utils.register_class(CustomDrawOperator)
    bpy.utils.register_class(ExplosionVariables)
    bpy.utils.register_class( ExplosionOperator )
    bpy.types.Scene.explosion_tool = bpy.props.PointerProperty(type=ExplosionVariables)
    bpy.utils.register_class(FancyPanel)
    bpy.utils.register_class(ExplosionPanel)
    bpy.utils.register_class(MessageBox)
    bpy.utils.register_class(RegisterHandlerOperator)
    bpy.utils.register_class(UnregisterHandlerOperator)
    bpy.utils.register_class(HandlerPanel)


def unregister():
    del bpy.types.Scene.explosion_tool
    del bpy.types.Scene.collection
    del bpy.types.Scene.amountcollection
    bpy.utils.unregister_class(ExplosionVariables)
    bpy.utils.unregister_class(ExplosionPanel)
    bpy.utils.unregister_class( ExplosionOperator )
    bpy.utils.unregister_class(CustomDrawOperator2)
    bpy.utils.unregister_class(CustomDrawOperator)
    bpy.utils.unregister_class(SceneItems)
    bpy.utils.unregister_class(AmountItems)
    bpy.utils.unregister_class(ForceLogFileOperator)
    bpy.utils.unregister_class(DebrisLogFileOperator)
    bpy.utils.unregister_class(AddButtonOperator)
    bpy.utils.unregister_class(MassButtonOperator)
    bpy.utils.unregister_class(BuildButtonOperator)
    bpy.utils.unregister_class(ButtonOperator)
    bpy.utils.unregister_class(DeleteOperator)
    bpy.utils.unregister_class(FancyPanel)
    bpy.utils.unregister_class(MessageBox)
    bpy.utils.unregister_class(RegisterHandlerOperator)
    bpy.utils.unregister_class(UnregisterHandlerOperator)
    bpy.utils.unregister_class(HandlerPanel)
    # Clear property
    if "force-contour" in bpy.app.driver_namespace:
        del bpy.app.driver_namespace["force-contour"]
    if "force-contour-dict" in bpy.app.driver_namespace:
        del bpy.app.driver_namespace["force-contour-dict"]
    if "explosion" in bpy.app.driver_namespace:
        del bpy.app.driver_namespace["explosion"]
    if "start-frame" in bpy.app.driver_namespace:
        del bpy.app.driver_namespace["start-frame"]

 
if __name__ == "__main__":
    try: unregister()
    except: pass
    register()